--
-- Dummy database upgrade to add MessageSearchTable, whose parameters depend on
-- things we need at run-time.  See src/engine/imap-db/imap-db-database.vala in
-- post_upgrade() for the code that runs the upgrade.
--
