/* Copyright 2016 Software Freedom Conservancy Inc.
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

/**
 * With GtkStock deprecated in GTK+ 3.10, these strings offer replacements for commonly-needed
 * text labels.
 *
 * Plain text use all-caps constant names and an underscore indicating where in the English text the
 * mnemonic lies.  This can be used to ensure that the mnemonic doesn't interfere with other custom
 * strings in the grouping.
 */

namespace Stock {

public const string _OK = _("_OK");
public const string _CANCEL = _("_Cancel");

public const string _ABOUT = _("_About");
public const string _ADD = _("_Add");
public const string _CLOSE = _("_Close");
public const string _DISCARD = _("_Discard");
public const string _HELP = _("_Help");
public const string _OPEN_BUTTON = _("_Open");
public const string _PREFERENCES = _("_Preferences");
public const string _PRINT_MENU = _("_Print…");
public const string _QUIT = _("_Quit");
public const string _REMOVE = _("_Remove");
public const string _SAVE = _("_Save");
public const string _KEEP = _("_Keep");

}

