/* Copyright 2016 Software Freedom Conservancy Inc.
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

/**
 * An error that is thrown when manipulating RFC 822 objects and data.
 */
public errordomain Geary.RFC822.Error {
    INVALID,
    NOT_FOUND,
    FAILED
}
