/* Copyright 2016 Michael Gratton <mike@vee.net>
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later). See the COPYING file in this distribution.
 */

/**
 * The primary namespace for the Geary email library.
 *
 * For the main entry point, see {@link Geary.Engine}.
 */
namespace Geary {

    // This source file exists purely as a place for the namespace's
    // documentation comment.

}
